import multiprocessing
import os
import platform
import shutil
import subprocess
from contextlib import contextmanager
from pathlib import Path
from typing import List

from Cython.Build import cythonize
from Cython.Distutils.build_ext import new_build_ext as cython_build_ext
from setuptools import Extension, Distribution


@contextmanager
def changedir(path):
    save_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(save_dir)

@contextmanager
def with_patches():
    patches = sorted([
        os.path.abspath(patch)
        for patch in Path("patches").iterdir()
        if patch.is_file() and patch.suffix == ".patch"
    ])
    has_git = shutil.which("git") is not None
    has_git_dir = Path(".git").exists() and Path(".git").is_dir()
    use_git = has_git and has_git_dir
    with changedir("bwa"):
        for patch in patches:
            if use_git:
                retcode = subprocess.call(f"git apply {patch}", shell=True)
            else:
                retcode = subprocess.call(f"patch -p1 < {patch}", shell=True)
            if retcode != 0:
                raise RuntimeError(f"Failed to apply patch {patch}")
    try:
        yield
    finally:
        if use_git:
            commands = ["git submodule deinit -f .", "git submodule update --init"]
            for command in commands:
                retcode = subprocess.call(command, shell=True)
                if retcode != 0:
                    raise RuntimeError(f"Failed to reset submodules: {command}")

compiler_directives = {
    "language_level": "3",
    'embedsignature': True,
}
SOURCE_DIR = Path("pybwa")
BUILD_DIR = Path("cython_build")
compile_args = []
link_args = []
include_dirs = ["bwa", "pybwa"]
libraries = ['m', 'z', 'pthread']
if platform.system() == 'Linux':
    libraries.append("rt")
library_dirs=['pybwa', 'bwa']
extra_objects = []
define_macros = [("HAVE_PTHREAD", None), ("USE_MALLOC_WRAPPERS", None)]
h_files = []
c_files = []

exclude_files = {
    "pybwa": ["libbwaaln.c", "libbwaindex.c", "libbwamem.c"],
    "bwa": ['example.c', 'main.c']
}
for root_dir in library_dirs:
    h_files.extend(str(x) for x in Path(root_dir).rglob("*.h"))
    c_files.extend(str(x) for x in Path(root_dir).rglob("*.c") if x.name not in exclude_files[root_dir])

# Check if we should build with linetracing for coverage
build_with_coverage = os.environ.get("CYTHON_TRACE", "false").lower() in ("1", "true", '"true"')
if build_with_coverage:
    compiler_directives["linetrace"] = True
    compiler_directives['emit_code_comments'] = True
    define_macros.extend([('CYTHON_TRACE', 1), ('CYTHON_TRACE_NOGIL', 1), ('DCYTHON_USE_SYS_MONITORING', 0)])
    BUILD_DIR = Path(".")  # the compiled .c files need to be next to the .pyx files for coverage

if platform.system() != 'Windows':
    compile_args.extend([
        "-Wno-unused-result",
        "-Wno-unreachable-code",
        "-Wno-single-bit-bitfield-constant-conversion",
        "-Wno-deprecated-declarations",
        "-Wno-unused",
        "-Wno-strict-prototypes",
        "-Wno-sign-compare",
        "-Wno-error=declaration-after-statement"
    ])

libbwaindex_module = Extension(
    name='pybwa.libbwaindex',
    sources=['pybwa/libbwaindex.pyx'] + c_files,
    depends=h_files,
    extra_compile_args=compile_args,
    extra_link_args=link_args,
    extra_objects=extra_objects,
    include_dirs=include_dirs,
    language='c',
    libraries=libraries,
    library_dirs=library_dirs,
    define_macros=define_macros
)

libbwaaln_module = Extension(
    name='pybwa.libbwaaln',
    sources=['pybwa/libbwaaln.pyx'] + c_files,
    depends=h_files,
    extra_compile_args=compile_args,
    extra_link_args=link_args,
    extra_objects=extra_objects,
    include_dirs=include_dirs,
    language='c',
    libraries=libraries,
    library_dirs=library_dirs,
    define_macros=define_macros
)

libbwamem_module = Extension(
    name='pybwa.libbwamem',
    sources=['pybwa/libbwamem.pyx'] + c_files,
    depends=h_files,
    extra_compile_args=compile_args,
    extra_link_args=link_args,
    extra_objects=extra_objects,
    include_dirs=include_dirs,
    language='c',
    libraries=libraries,
    library_dirs=library_dirs,
    define_macros=define_macros
)


def cythonize_helper(extension_modules: List[Extension]) -> List[Extension]:
    """Cythonize all Python extensions"""

    return cythonize(
        module_list=extension_modules,

        # Don't build in source tree (this leaves behind .c files)
        build_dir=BUILD_DIR,

        # Don't generate an .html output file. Would contain source.
        annotate=False,

        # Parallelize our build
        nthreads=multiprocessing.cpu_count() * 2,

        # Compiler directives (e.g. language, or line tracing for coverage)
        compiler_directives=compiler_directives,

        # (Optional) Always rebuild, even if files untouched
        force=True,
    )

CLASSIFIERS = '''
Development Status :: 4 - Beta
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved
Programming Language :: Python
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
'''


def build():
    # apply patches to bwa, then revert them after
    with with_patches():
        # Collect and cythonize all files
        extension_modules = cythonize_helper([
            libbwaindex_module,
            libbwaaln_module,
            libbwamem_module
        ])

        # Use Setuptools to collect files
        distribution = Distribution({
            "name": "pybwa",
            'version': '0.0.1',
            'description': 'Python bindings for BWA',
            'long_description': __doc__,
            'long_description_content_type': 'text/x-rst',
            'author': 'Nils Homer',
            'author_email': 'nils@fulcrumgenomics.com',
            'license': 'MIT',
            'platforms': ['POSIX', 'UNIX', 'MacOS'],
            'classifiers': [_f for _f in CLASSIFIERS.split('\n') if _f],
            'url': 'https://github.com/fulcrumgenomics/pybwa',
            'packages': ['pybwa', 'pybwa.include.bwa', 'pybwa.include.patches'],
            'package_dir': {'pybwa': 'pybwa', 'pybwa.include.bwa': 'bwa', 'pybwa.include.patches': 'patches' },
            'package_data': {'': ['*.pxd', '*.h', '*.c', 'py.typed', '*.pyi', '*.patch'], },
            "ext_modules": extension_modules,
            "cmdclass": {
                "build_ext": cython_build_ext,
            },
        })

        # Grab the build_ext command and copy all files back to source dir.
        # Done so Poetry grabs the files during the next step in its build.
        build_ext_cmd = distribution.get_command_obj("build_ext")
        build_ext_cmd.ensure_finalized()
        # Set the value to 1 for "inplace", with the goal to build extensions
        # in build directory, and then copy all files back to the source dir
        # (under the hood, "copy_extensions_to_source" will be called after
        # building the extensions). This is done so Poetry grabs the files
        # during the next step in its build.
        build_ext_cmd.inplace = 1
        build_ext_cmd.run()


if __name__ == "__main__":
    build()



from pathlib import Path
from typing import Optional

import pytest
from fgpyo.sequence import reverse_complement
from pysam import FastxRecord

from pybwa import BwaAln
from pybwa import BwaAlnOptions
from pybwa import BwaIndex
from pybwa.libbwamem import BwaMem
from pybwa.libbwamem import BwaMemOptions


@pytest.fixture()
def ref_fasta() -> Path:
    cur_dir = Path(__file__).parent
    fasta: Path = cur_dir / "data" / "e_coli_k12.fasta"
    return fasta


def test_bwa_index_build(ref_fasta: Path, tmp_path_factory: pytest.TempPathFactory) -> None:
    tmp_dir = Path(str(tmp_path_factory.mktemp("test_bwa_index_build")))
    prefix = tmp_dir / ref_fasta.name

    # Build the index
    BwaIndex.index(fasta=ref_fasta, prefix=prefix)

    # Check the files exist
    for suffix in [".bwt", ".sa", ".ann", ".pac"]:
        assert prefix.with_suffix(prefix.suffix + suffix).exists()
    assert prefix.with_suffix(".dict").exists()

    # Load it
    BwaIndex(prefix=prefix)


@pytest.fixture(scope="function")
def temp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    return tmp_path_factory.mktemp("test_vcf")


def test_bwa_index(ref_fasta: Path) -> None:
    BwaIndex(prefix=ref_fasta)


def test_bwaaln_options() -> None:
    BwaAlnOptions()
    # TODO: test setting individual options...


@pytest.fixture()
def fastx_record() -> FastxRecord:
    sequence = "gttacctgccgtgagtaaattaaaattttattgacttaggtcactaaatactttaaccaatataggcatagcgcacagac"
    return FastxRecord(name="test", sequence=sequence)


def test_bwaaln(ref_fasta: Path, fastx_record: FastxRecord) -> None:
    opt = BwaAlnOptions()
    bwa = BwaAln(prefix=ref_fasta)

    revcomp_seq = None if not fastx_record.sequence else reverse_complement(fastx_record.sequence)
    revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)

    recs = bwa.align(opt=opt, queries=[fastx_record, revcomp_record])
    assert len(recs) == 2
    rec = recs[0]
    assert rec.query_name == "test"
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.reference_start == 80
    assert rec.is_forward
    assert rec.cigarstring == "80M"

    rec = recs[1]
    assert rec.query_name == "revcomp"
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.reference_start == 80
    assert rec.is_reverse
    assert rec.cigarstring == "80M"

    # NB: XN, XA not generated for these records
    expected_tags = ["NM", "X0", "X1", "XM", "XO", "XG", "MD", "HN"]
    for rec in recs:
        for tag in expected_tags:
            assert rec.has_tag(tag), f"Missing tag {tag} in: {rec}"


def test_bwaaln_threading(ref_fasta: Path, fastx_record: FastxRecord) -> None:
    opt = BwaAlnOptions(threads=2)
    bwa = BwaAln(prefix=ref_fasta)
    revcomp_seq = None if not fastx_record.sequence else reverse_complement(fastx_record.sequence)
    revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)

    queries = [fastx_record if i % 2 == 0 else revcomp_record for i in range(100)]
    recs = bwa.align(opt=opt, queries=queries)
    assert len(recs) == len(queries)
    for i, rec in enumerate(recs):
        if i % 2 == 0:
            assert rec.query_name == "test"
            assert rec.is_forward
        else:
            assert rec.query_name == "revcomp"
            assert rec.is_reverse
        assert not rec.is_paired
        assert not rec.is_read1
        assert not rec.is_read2
        assert rec.reference_start == 80
        assert rec.cigarstring == "80M"


@pytest.mark.parametrize(
    "max_gap_extensions,is_mapped,deletion_length",
    [
        (None, True, 4),  # 4bp deletions allowed (max diff is 4)
        (None, False, 5),  # 5bp deletions disallowed (max diff is 4)
        (-1, True, 1),  # 1bp deletions allowed
        (-1, False, 2),  # 2bp deletions disallowed
        (0, True, 1),  # 1bp deletions allowed
        (0, False, 2),  # 2bp deletions disallowed
        (4, True, 5),  # 5bp deletion allowed (1 open + 4 extensions)
        (4, False, 6),  # 6bp deletion disallowed (1 open + 4 extensions)
    ],
)
def test_bwaaln_with_deletion(
    ref_fasta: Path,
    fastx_record: FastxRecord,
    deletion_length: int,
    max_gap_extensions: Optional[int],
    is_mapped: bool,
) -> None:
    bwa = BwaAln(prefix=ref_fasta)

    # Create a deletion
    assert fastx_record.sequence is not None
    sequence = fastx_record.sequence[:40] + fastx_record.sequence[40 + deletion_length :]
    fastx_record = FastxRecord(name=fastx_record.name, sequence=sequence)

    opt = BwaAlnOptions(max_gap_extensions=max_gap_extensions)
    recs = bwa.align(opt=opt, queries=[fastx_record])
    assert len(recs) == 1
    rec = recs[0]
    assert rec.query_name == "test"
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2

    if is_mapped:
        assert rec.reference_start == 80
        assert rec.is_forward
        assert rec.cigarstring == f"40M{deletion_length}D{40-deletion_length}M"
    else:
        assert rec.reference_start == -1, fastx_record.sequence
        assert rec.is_unmapped


def test_bwa_aln_map_one_multi_mapped_max_hits_one(ref_fasta: Path) -> None:
    """Tests that a query that returns too many hits (>max_hits) returns the number of hits but
    not the list of hits themselves."""
    opt = BwaAlnOptions(
        threads=2,
        max_hits=1,
        max_mismatches=3,
        max_mismatches_in_seed=3,
        max_gap_opens=0,
        max_gap_extensions=-1,
        min_indel_to_end_distance=3,
        seed_length=20,
        find_all_hits=True,
        with_md=True,
    )
    bwa = BwaAln(prefix=ref_fasta)
    queries = [FastxRecord(name="NA", sequence="TTTTT")]
    recs = bwa.align(opt=opt, queries=queries)
    assert len(recs) == 1
    rec = recs[0]
    assert rec.has_tag("HN"), str(rec)
    assert rec.get_tag("HN") == 3269888


def test_bwamem_options() -> None:
    # default options
    options = BwaMemOptions()
    assert options.min_seed_len == 19

    # finalize, returning a copy
    copy = options.finalize(copy=True)
    assert copy.min_seed_len == 19
    assert copy.finalized
    assert not options.finalized

    # update min seed len
    options.min_seed_len = 20
    assert options.min_seed_len == 20

    # finalize, returning a copy with a new value
    copy = options.finalize(copy=True)
    assert copy.min_seed_len == 20
    assert copy.finalized
    assert not options.finalized

    # finalize, returning itself finalized
    options.finalize()
    assert options.finalized
    assert options.min_seed_len == 20  # type:ignore[unreachable]

    # raise an exception if we try to set a new value
    with pytest.raises(AttributeError):
        options.min_seed_len = 19


def test_bwamem(ref_fasta: Path, fastx_record: FastxRecord) -> None:
    opt = BwaMemOptions(with_xr_tag=True)
    bwa = BwaMem(prefix=ref_fasta)

    revcomp_seq = None if not fastx_record.sequence else reverse_complement(fastx_record.sequence)
    revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)

    recs_of_recs = bwa.align(opt=opt, queries=[fastx_record, revcomp_record])
    assert len(recs_of_recs) == 2

    assert len(recs_of_recs[0]) == 1
    rec = recs_of_recs[0][0]
    assert rec.query_name == "test"
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.reference_start == 80
    assert rec.is_forward
    assert rec.cigarstring == "80M"

    assert len(recs_of_recs[1]) == 1
    rec = recs_of_recs[1][0]
    assert rec.query_name == "revcomp"
    assert not rec.is_paired
    assert not rec.is_read1
    assert not rec.is_read2
    assert rec.reference_start == 80
    assert rec.is_reverse
    assert rec.cigarstring == "80M"
    # TODO: test multi-mapping, reverse strand, etc

    # NB: XA amd XB not generated for these records
    expected_tags = ["NM", "MD", "AS", "XS", "XR"]
    for recs in recs_of_recs:
        for rec in recs:
            for tag in expected_tags:
                assert rec.has_tag(tag), f"Missing tag {tag} in: {rec}"


def test_bwamem_threading(ref_fasta: Path, fastx_record: FastxRecord) -> None:
    opt = BwaMemOptions(threads=2)
    bwa = BwaMem(prefix=ref_fasta)

    revcomp_seq = None if not fastx_record.sequence else reverse_complement(fastx_record.sequence)
    revcomp_record = FastxRecord(name="revcomp", sequence=revcomp_seq)

    queries = [fastx_record if i % 2 == 0 else revcomp_record for i in range(100)]
    list_of_recs = bwa.align(opt=opt, queries=queries)
    assert len(list_of_recs) == len(queries)
    for i, recs in enumerate(list_of_recs):
        assert len(recs) == 1
        rec = recs[0]
        if i % 2 == 0:
            assert rec.query_name == "test"
            assert rec.is_forward
        else:
            assert rec.query_name == "revcomp"
            assert rec.is_reverse
        assert not rec.is_paired
        assert not rec.is_read1
        assert not rec.is_read2
        assert rec.reference_start == 80
        assert rec.cigarstring == "80M"
